from __future__ import print_function
import signal
import pyuv
import os


def watch(triggers, directory=os.getcwd(), lockfiles=None):
    """Initialize the watcher and run it."""
    Watcher(triggers, directory, lockfiles).run()


class Watcher(object):
    """Watches the filesystem and fires triggers using libuv."""

    def __init__(self, triggers, directory, lockfiles):
        self.directory = directory
        self.triggers = triggers
        self.lockfiles = [] if lockfiles is None else lockfiles
        self.change_queue = []
        self.event_handles = []
        self.task_queue = []
        self.signal_handler = None
        self.timer_handler = None
        self.triggered = False

    def empty_task_queue(self):
        """Is the task queue empty?"""
        return len(self.task_queue) == 0

    def first_task(self):
        """Return first task in the queue, or nothing if empty."""
        if self.empty_task_queue():
            return None
        else:
            return self.task_queue[0]

    def close_handles(self):
        """Close all I/O handles prior to exit,
           provided a non-ctrlc ignored task is in progress."""
        if self.first_task() is None or not self.first_task().trigger.ignore_ctrlc:
            for task in self.task_queue:
                task.stop()
            self.signal_handler.close()
            self.timer_handler.close()
            for event_handle in self.event_handles:
                event_handle.close()

    def read_callback(self, handle, filename, events, error):
        """Callback every time something is modified in the repository."""
        fullpath = os.path.realpath(str(handle.path) + os.sep + filename)
        relative_path = fullpath.replace(os.path.realpath(self.directory) + os.sep, "")

        if os.path.exists(fullpath):
            self.change_queue.append(relative_path)

        for lockfile in self.lockfiles:
            if os.path.exists(self.directory + os.sep + lockfile):
                return

        for trigger in self.triggers:
            task = trigger.hit_with(self.change_queue)
            if task is not None and task not in self.task_queue[1:]:
                self.task_queue.append(task)
                self.triggered = True
        self.change_queue = []

    def poll_callback(self, timer_handle):
        """Handle running tasks."""
        if len(self.task_queue) > 0:
            firsttask = self.task_queue[0]
            if self.triggered and len(self.task_queue) > 1:
                firsttask.attempt_stop()

            self.triggered = False
            firsttask.start()
            if firsttask.is_done():
                del self.task_queue[0]

                if len(self.task_queue) == 0:
                    print("PATROL: ALL TASKS COMPLETE")



    def run(self):
        """Run the watcher."""
        print("PATROL: Started")
        loop = pyuv.Loop.default_loop()

        # Attach a handler for each subdirectory underneath
        for subdirectory in [os.path.realpath(x[0]) for x in os.walk(self.directory)]:
            event_handle = pyuv.fs.FSEvent(loop)
            event_handle.start(subdirectory, 0, self.read_callback)
            self.event_handles.append(event_handle)

        # Attach a handler for CTRL-C
        self.signal_handler = pyuv.Signal(loop)
        self.signal_handler.start(lambda handle, signum: self.close_handles(), signal.SIGINT)

        # Attach a handler to start/stop/kill running tasks.
        self.timer_handler = pyuv.Timer(loop)
        self.timer_handler.start(self.poll_callback, 0.05, 0.05)

        # Fire all "fire on initialization" triggers.
        for trigger in self.triggers:
            if trigger.fire_on_initialization:
                self.task_queue.append(trigger.fire([]))

        loop.run()
        print("PATROL: Stopped")



from __future__ import print_function
from patrol.task import Task
import fnmatch



class Trigger(object):
    """Some conditions and a python method that yields a task."""

    def __init__(self,
                 method,
                 includes,
                 excludes=None,
                 args=None,
                 reaper=None,
                 ignore_ctrlc=False,
                 fire_on_initialization=False):

        self.method = method
        self.includes = includes
        self.args = [] if args is None else args
        self.excludes = [] if excludes is None else excludes
        self.reaper = reaper
        self.ignore_ctrlc = ignore_ctrlc
        self.fire_on_initialization = fire_on_initialization

    def _match(self, filenames):
        """Return True if trigger matches one of the specified filenames."""
        currently_matching = False

        for filename in filenames:
            for include in self.includes:
                if fnmatch.fnmatch(filename, include):
                    currently_matching = True

            if self.excludes is not None:
                for exclude in self.excludes:
                    if fnmatch.fnmatch(filename, exclude):
                        currently_matching = False
        return currently_matching

    def hit_with(self, filenames):
        """Return new task if filenames match trigger, else return None."""
        if self._match(filenames):
            return self.fire(filenames)
        else:
            return None

    def fire(self, filenames):
        """Fire trigger and return task."""
        if filenames == []:
            print("PATROL: Task '{}' triggered by changes to '{}'".format(
                self.method.__name__, ', '.join(filenames)
            ))
        else:
            print("PATROL: Task '{}' triggered.".format(self.method.__name__))
        return Task(self, filenames)



import subprocess
import unittest
import shutil
import time
import sys
import os


class PatrolTestCase(unittest.TestCase):
    """Run a patrol script that mimics a running build and a test."""
    TEST_DIR = os.path.dirname(os.getcwd() + os.sep + str(__file__))

    def setUp(self):
        # Change directory to the folder this file is in
        os.chdir(self.TEST_DIR)
        self.clean_output()

    def wait_for_file_to_exist(self, filename, timeout=2):
        for _ in range(0, int(timeout * 100)):
            time.sleep(0.01)
            if os.path.exists(filename):
                return
        raise AssertionError("{} filename never appeared.".format(filename))

    def clean_output(self):
        shutil.rmtree("output", ignore_errors=True)
        os.makedirs("output")

    def touch(self, filename):
        """Platform independent equivalent of UNIX touch cmd."""
        with open(filename, "a"):
            os.utime(filename, None)

    def test_fire_on_initialization(self):
        """Test that triggers got fired without a change."""
        patrol_proc = subprocess.Popen([sys.executable, "patrol1.py",])

        self.wait_for_file_to_exist("output/build_started", timeout=0.5)
        self.wait_for_file_to_exist("output/build_finished", timeout=1)
        self.wait_for_file_to_exist("output/test_started", timeout=0.5)
        self.wait_for_file_to_exist("output/test_finished", timeout=2)

        patrol_proc.kill()

    def test_reaper(self):
        patrol_proc = subprocess.Popen([sys.executable, "patrol1.py",])

        time.sleep(3.5)
        self.clean_output()
        self.touch("data/a")
        self.wait_for_file_to_exist("output/build_started", timeout=0.5)
        self.wait_for_file_to_exist("output/build_finished", timeout=1)
        self.wait_for_file_to_exist("output/test_started", timeout=0.5)
        self.touch("data/a")
        self.clean_output()
        self.wait_for_file_to_exist("output/build_started", timeout=0.5)
        self.wait_for_file_to_exist("output/build_finished", timeout=1)
        self.wait_for_file_to_exist("output/test_started", timeout=0.5)
        self.assertFalse(os.path.exists("output/test_finished"))

        patrol_proc.kill()




if __name__ == "__main__":
    unittest.main()



import multiprocessing


class Task(object):
    """A task to be completed in a method."""

    def __init__(self, trigger, filenames):
        self.trigger = trigger
        self.args = tuple([filenames] + list(self.trigger.args))
        self.process = multiprocessing.Process(target=self.run_method)
        self.started = False

    def __eq__(self, other):
        """Two tasks are the same if the method being run is the same."""
        return self.trigger.method == other.trigger.method

    def run_method(self):
        """Run trigger method with args (this should be run in a different process)."""
        self.trigger.method(self.args)

    def start(self):
        """Start task if not already started."""
        if not self.started:
            self.started = True
            self.process.start()

    def is_done(self):
        """Has the task finished?"""
        return self.started and not self.process.is_alive()

    def attempt_stop(self):
        """Send reaper if there is one, assuming the task was started."""
        if self.trigger.reaper is not None and self.started:
            self.stop()

    def stop(self):
        """Send the reaper after the task."""
        if self.process.is_alive():
            self.trigger.reaper.reap(self.trigger.method.__name__, self.process)



from __future__ import print_function
from signal import SIGTERM
import psutil
import time


class Reaper(object):
    """Object responsible for stopping processes - softly or violently."""

    def __init__(self, signal=SIGTERM, timeout=10):
        self.signal = signal
        self.timeout = timeout

    def reap(self, name, process):
        print("PATROL: Stopping task '{}'".format(name))
        proc_list = psutil.Process(process.pid).get_children(recursive=True)
        for childproc in proc_list:
            try:
                childproc.send_signal(self.signal)
            except psutil.NoSuchProcess:
                pass

        for _ in range(0, int(self.timeout * 10)):
            time.sleep(0.1)
            still_alive = False

            for childproc in proc_list:
                if childproc.is_running():
                    still_alive = True

            if not still_alive:
                break

        if still_alive:
            for proc in proc_list:
                if proc.is_running():
                    full_name = ' '.join(proc.cmdline).strip()
                    print("PATROL: Killing process '{}'".format(full_name))
                    proc.kill()
        else:
            print("PATROL: '{}' finished cleanly.".format(name))

        process.terminate()



from os import path, chdir, getcwd, utime, sep
import patrol
import shutil
import time
import os

# Change directory to the directory this file is in
chdir(path.dirname(getcwd() + sep + __file__))

def touch(filename):
    """Platform independent equivalent of UNIX touch cmd."""
    with open(filename, "a"):
        utime(filename, None)

def build(filenames):
    touch("output/build_started")
    time.sleep(1)
    touch("output/build_finished")

def run_test(filenames):
    touch("output/test_started")
    time.sleep(2)
    touch("output/test_finished")

patrol.watch([
        patrol.Trigger(
            build,
            includes=["data/*", ],
            excludes=['data/exclude/*', 'output/*', ],
            fire_on_initialization=True, # When patrol.watch() is first run, initiate trigger.
        ),
        patrol.Trigger(
            run_test,
            includes=["data/*", ],
            excludes=['data/exclude/*', 'output/*', ],
            reaper=patrol.Reaper(),   # If triggered while method is in progress, it will stop it and start again.
            fire_on_initialization=True,
        ),
    ],
    directory=getcwd(),
    lockfiles=["lock", ],
)



from setuptools.command.install import install
from setuptools import setup, find_packages
import codecs
import sys
import os


def read(*parts):
    # intentionally *not* adding an encoding option to open
    # see here: https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(os.path.abspath(os.path.dirname(__file__)), *parts), 'r').read()

setup(name="patrol",
      version="0.3",
      description="Trigger custom commands from filesystem events.",
      long_description=read('README.rst'),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Topic :: Software Development :: Build Tools',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.1',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
      ],
      keywords='development environment tool pyuv epoll build',
      author='Colm O\'Connor',
      author_email='colm.oconnor.github@gmail.com',
      packages=find_packages(exclude=["tests/", ]),
      url='https://github.com/crdoconnor/patrol',
      license='MIT',
      install_requires=['pyuv', 'psutil', ],
      package_data={},
      zip_safe=False,)
